#include <string.h>
#include <gtk/gtk.h>
#include <filter.h>
#include <jmp-config.h>

/* The filter string for class names */
static GString *classfilter;
static GtkEntry *classfilter_entry;

/* Dialog for filter options */
static GtkWidget *dialog;
static GtkWidget *filter_mode_check_boxes[4];

/* One of FILTER_MATCH_ALL, FILTER_EXACT, FILTER_PACKAGE or FILTER_RECURSIVE */
static int filter_mode = FILTER_MATCH_ALL;
/* Filter mode visible in the gui */
static int filter_mode_selected;

int filter_class_name (const char *name)
{
    int r;

    if (filter_mode == FILTER_MATCH_ALL) 
	return TRUE;

    if (filter_mode == FILTER_MATCH_CLASS)
	return !strcmp (name, classfilter->str);

    if (filter_mode == FILTER_MATCH_RECURSIVE)
	return !strncmp(name, classfilter->str, classfilter->len);

    /* FILTER_MATCH_PACKAGE */
    r = strncmp (name, classfilter->str, classfilter->len);
    if (r)
	return FALSE;
    else
	return (!index (name + classfilter->len, '.'));
  
}

/* Set the filter. 
 * @param filtertype one of FILTER_MATCH_*
 * @param filter the string to filter on ("rabbit.proxy").
 */
void filter_set_filter (int filtertype, const char *filter) {
    filter_mode = filtertype;
    if (classfilter)
	g_string_free (classfilter, TRUE);
    classfilter = g_string_new (filter);
}

/* Called when OK or Quit button of the options dialog is pressed */
void filter_close_dialog(GtkWidget *widget, gpointer data)
{
    /* OK? */
    if (data) {
	g_string_free (classfilter, TRUE);
	classfilter = g_string_new (gtk_entry_get_text (classfilter_entry));
	filter_mode = filter_mode_selected;
	if (filter_mode == FILTER_MATCH_PACKAGE
	    || filter_mode == FILTER_MATCH_RECURSIVE) {
	    if (classfilter->len == 0 || classfilter->str[classfilter->len-1] != '.') {
		g_string_append_c (classfilter, '.');
	    }
	}
    }
    gtk_widget_hide (dialog);
}

void filter_mode_changed(GtkWidget *w,gpointer fm)
{
    if (GTK_TOGGLE_BUTTON (w)->active)
	filter_mode_selected = (int)fm;
}

/* Opens a dialog to edit filter options */
void filter_edit_options (GtkObject *ignored)
{
    GtkTooltips *tooltips = gtk_tooltips_new ();

    /* Window already initialized? */
    if (dialog == NULL) {
	GtkWidget *b = gtk_vbox_new (TRUE,0), *b2;
	GtkWidget *w;
	int i;
	char *filter_mode_text[4] = {_("match class"),_("match package"),_("match recursive"),
				     _("match all")};
	char *filter_mode_tooltip[4] = {_("Matches a specific class: \"package.subpackage.MyClass\""),
					_("Matches a package: \"package.subpackage\""),
					_("Matches a package and all subpcagkes of it: \"package\""),
					_("Matches everything") };
	
	if (!classfilter)
	    classfilter = g_string_new ("");

	dialog = gtk_window_new (GTK_WINDOW_DIALOG);
	gtk_window_set_title (GTK_WINDOW (dialog),
			      _("Java Memory Profiler - Filter Options"));
	gtk_window_set_modal (GTK_WINDOW (dialog),TRUE);
	gtk_container_set_border_width (GTK_CONTAINER (dialog),10);

	/* Text field for the filter string */
	w = gtk_label_new (_("Enter class or package name:"));
	gtk_label_set_justify (GTK_LABEL (w),GTK_JUSTIFY_LEFT);
	gtk_widget_show (w);
	b2 = gtk_hbox_new (FALSE,0);
	gtk_box_pack_start (GTK_BOX (b2),w,FALSE,FALSE,0);
	gtk_widget_show (b2);
	gtk_box_pack_start (GTK_BOX (b),b2,FALSE,FALSE,0);
	classfilter_entry = GTK_ENTRY (gtk_entry_new ());
	gtk_widget_show (GTK_WIDGET (classfilter_entry));
	gtk_box_pack_start (GTK_BOX (b),GTK_WIDGET (classfilter_entry),FALSE,FALSE,0);

	/* Radio buttons for match type */
	for  (i = 0; i < 4; i++) {
	    filter_mode_check_boxes[i] = 
		w = gtk_radio_button_new_with_label
		(i != 0 ? gtk_radio_button_group (GTK_RADIO_BUTTON (w)) : NULL,
		 filter_mode_text[i]);
	    gtk_tooltips_set_tip(tooltips, w, filter_mode_tooltip[i], NULL);
	    gtk_widget_show (w);
	    gtk_box_pack_start (GTK_BOX (b), w, FALSE,FALSE,0);
	    gtk_signal_connect (GTK_OBJECT (w), "toggled",
				GTK_SIGNAL_FUNC (filter_mode_changed), (gpointer) i);
	}


	/* Ok and Quit button */
	b2 = gtk_hbutton_box_new ();
	w = gtk_button_new_with_label (_("OK"));
	gtk_widget_show (w);
	gtk_signal_connect (GTK_OBJECT (w),"clicked",
			    GTK_SIGNAL_FUNC (filter_close_dialog),
			    (gpointer)TRUE);
	gtk_box_pack_start (GTK_BOX (b2), w, FALSE,FALSE,0);
	w = gtk_button_new_with_label (_("Quit"));
	gtk_widget_show (w);
	gtk_signal_connect (GTK_OBJECT (w), "clicked",
			    GTK_SIGNAL_FUNC (filter_close_dialog),
			    (gpointer)FALSE);
	gtk_box_pack_start (GTK_BOX (b2), w, FALSE,FALSE,0);
	gtk_widget_show (b2);
	gtk_box_pack_start (GTK_BOX (b), b2, FALSE,FALSE,0);
    
	gtk_widget_show (b);
	gtk_container_add (GTK_CONTAINER (dialog), b);
    
    }
    gtk_entry_set_text (classfilter_entry, classfilter->str);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (filter_mode_check_boxes[filter_mode]), TRUE);
    gtk_widget_show (dialog);
}

