/* Handle a hash table.
 */

#include <stdio.h>
#include <stdlib.h>
#include "hash.h"

struct hashnode
{
    void              *el;
    struct hashnode   *next;
};
typedef struct hashnode    hashnode;

struct hashtab
{
    size_t        size;
    size_t        cardinal;
    hash_f        hashfun;
    hash_cmp_f    cmp;
    hashnode    **vec;
    hashnode     *next_free;
};

#define HASH(tab, el) \
  ((tab)->hashfun (el, (tab)->size))

/** Good sizes of the hash table.
 *  primes, doubling in size so if we need to restructure 
 *  we know that (aprox) half of it is used...
 */
static unsigned long   primes[] =
{ 11UL, 23UL, 47UL, 97UL, 197UL, 307UL,
  617UL, 1237UL, 2477UL, 4957UL, 9923UL,
  19853UL, 39709UL, 79423UL, 158849UL,
  317701UL, 635413UL, 1270849UL,
  2541701UL, 5083423UL };

static size_t
make_prime (
    size_t   num)
{
    int      i;

    for (i = 0;
         i < sizeof (primes) / sizeof (int)
	     && primes[i] < num;
         ++i)
        ;
    if (i < sizeof (primes) / sizeof (int))
        return primes[i];
    else
        return num;
}



hashtab *
hash_new (
    size_t         size,
    hash_f         hashfun,
    hash_cmp_f     cmp)
{
    hashtab       *htab;
    size_t         i;

    size = make_prime (size);
    htab = malloc (sizeof (*htab));
    if (htab == NULL)
        return NULL;
    htab->vec = malloc (size * sizeof (hashnode *));
    if (htab->vec == NULL)
    {
        free (htab);
        return NULL;
    }
    htab->size = size;
    htab->cardinal = 0;
    htab->hashfun = hashfun;
    htab->cmp = cmp;
    for (i = 0; i < size; ++i)
        htab->vec[i] = NULL;

    htab->next_free = NULL;
    return htab;
}

static hashnode* new_node(hashtab *table) {
    if (table->next_free) {
	hashnode *toreturn = table->next_free;
	table->next_free = toreturn->next;
	return toreturn;
    }
    return (hashnode*) malloc (sizeof (hashnode));
}

static void free_node(hashtab *table,hashnode *node)
{
    node->next = table->next_free;
    table->next_free = node;
}

void
hash_free (
    hashtab    *htab)
{
    size_t      i;
    hashnode   *cur;
    hashnode   *next;

    for (i = 0; i < htab->size; ++i)
    {
        cur = htab->vec[i];
        while (cur != NULL)
        {
            next = cur->next;
            free (cur);
            cur = next;
        }
    }
    cur = htab->next_free;
    while (cur != NULL)
    {
	next = cur->next;
	free (cur);
	cur = next;
    }
    free (htab->vec);
    free (htab);
}



static void
restructure (
    hashtab    *htab)
{
    hashtab    *ntab = NULL;
    hashnode   *cur;
    hashnode  **tmpvec;
    size_t      i;
    int         factor;
    size_t      tmpsize;

    for (factor = 8; ntab == NULL && factor > 1; factor /= 2)
        ntab = hash_new (factor * htab->size,
                         htab->hashfun,
                         htab->cmp);
    if (ntab == NULL)
        return;

    for (i = 0; i < htab->size; ++i)
    {
        cur = htab->vec[i];
        while (cur != NULL)
        {
            if (hash_insert (cur->el, ntab) != 0)
            {
                hash_free (ntab);
                return;
            }
            cur = cur->next;
        }
    }
    tmpvec = htab->vec;
    tmpsize = htab->size;
    htab->vec = ntab->vec;
    htab->size = ntab->size;
    ntab->vec = tmpvec;
    ntab->size = tmpsize;
    hash_free (ntab);
}

int
hash_insert (
    void       *el,
    hashtab    *htab)
{
    size_t      i;
    hashnode   *hnode;

    if (htab == NULL) 
	return 0;
    
    ++htab->cardinal;
    if (htab->cardinal > htab->size * 10)
        restructure (htab);
    hnode = new_node(htab);
    if (hnode == NULL)
        return 1;
    
    i = HASH (htab, el);
    hnode->next = htab->vec[i];
    hnode->el = el;
    htab->vec[i] = hnode;
    return 0;
}

static hashnode **
find_elem (
    void        *el,
    hashtab     *htab)
{
    hashnode   **hnode_p;
    
    hnode_p = &htab->vec[HASH (htab, el)];
    while ((*hnode_p != NULL) && (htab->cmp ((*hnode_p)->el, el)))
        hnode_p = &(*hnode_p)->next;
    return hnode_p;
}

void *
hash_search (
    void        *el,
    hashtab     *htab)
{
    hashnode   **hnode_ptr;

    hnode_ptr = find_elem (el, htab);
    if (*hnode_ptr == NULL)
        return NULL;
    else
        return (*hnode_ptr)->el;
}

void *
hash_remove (
    void        *el,
    hashtab     *htab)
{
    hashnode   **hnode_ptr;
    hashnode    *remove;
    void        *ret_el;

    hnode_ptr = find_elem (el, htab);
    if (*hnode_ptr == NULL)
        return NULL;

    ret_el = (*hnode_ptr)->el;
    remove = *hnode_ptr;
    *hnode_ptr = (*hnode_ptr)->next;
    free_node (htab, remove);
    --htab->cardinal;

    return ret_el;
}

size_t
hash_cardinal (
    hashtab   *htab)
{
    return htab->cardinal;
}

void
hash_for_each (
    hash_iter_f    proc,
    hashtab       *htab)
{
    size_t         i;
    hashnode      *hnode;

    for (i = 0; i < htab->size; ++i)
        for (hnode = htab->vec[i];
             hnode != NULL;
             hnode = hnode->next)
        {
            proc (hnode->el);
        }
}

void
hash_for_each_with_arg (
    hash_iter_fa   proc,
    hashtab       *htab,
    void          *data)
{
    size_t         i;
    hashnode      *hnode;

    for (i = 0; i < htab->size; ++i)
        for (hnode = htab->vec[i];
             hnode != NULL;
             hnode = hnode->next)
        {
            proc (hnode->el, data);
        }
}
