#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <jni.h>
#include <jvmpi.h>
#include <jmp-config.h>

/*#include <assert.h>*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <jmp.h>
#include <hash.h>
#include <arena.h>
#include <cls.h>
#include <obj.h>
#include <method.h>
#include <jthread.h>
#include <timerstack.h>
#include <objectstore.h>
#include <ui.h>
#include <object_dump.h>
#include <heap_dump.h>
#include <instance_owners.h>

/* Compile with -DJMPDEBUG to get lots of traces, can be helpful sometimes. */
#ifndef JMPDEBUG
#define trace(a) 
#define trace2(a,b) 
#define trace3(a,b,c)
#define trace4(a,b,c,d)
#else
#define trace(a) fprintf (stderr, a); fflush (stderr)
#define trace2(a,b) fprintf (stderr, a, b); fflush (stderr)
#define trace3(a,b,c) fprintf (stderr, a, b, c); fflush (stderr)
#define trace4(a,b,c,d) fprintf (stderr, a, b, c, d); fflush (stderr)
#endif /* JMPDEBUG */

/** This is the interface into the jvm */
static JVMPI_Interface* jvmpi;

/** TODO: evaluate if we can have more locks and thus get faster operations. */
/** Threads and method lock */
static JVMPI_RawMonitor data_access_lock;
/** Object alloc/move/free lock */
static JVMPI_RawMonitor objects_access_lock;

/** are we showing any graphical user interface? */
static int doUI = 1;
/** are we profiling method calls? */
static int method_profiling = 1;
/** are we tracing objects? */
static int object_profiling = 1;
/** are we tracing monitors. */
static int monitor_profiling = 1;
/** do we allow dump on signal? */
static int dump_enabled = 0;


/** These are the hashes where we store our different objects. */
static hashtab* arenas;
static hashtab* classes;
static hashtab* objects;
static hashtab* methods;
static hashtab* threads;

/** This is our objectstore. */
static objectstore* osp;

/** The names of the primitive array types. */
static char* L = "[L";
static cls* LP = 0;
static char* Z = "[Z";
static cls* ZP = 0;
static char* B = "[B";
static cls* BP = 0;
static char* C = "[C";
static cls* CP = 0;
static char* S = "[S";
static cls* SP = 0;
static char* I = "[I";
static cls* IP = 0;
static char* J = "[J";
static cls* JP = 0;
static char* F = "[F";
static cls* FP = 0;
static char* D = "[D";
static cls* DP = 0;

/** Counters for methods. */
static long c_class_load = 0;
static long c_class_unload = 0;
static long c_object_alloc = 0;
static long c_object_move = 0;
static long c_object_free = 0;
static long c_thread_start = 0;
static long c_thread_end = 0;
static long c_method_entry = 0;
static long c_method_exit = 0;

/** Where should we dump data files? */
static char* dumpfile = NULL;

static int down = 0;   /** since we get multiple shutdowns... */

/** Get the lock on the object monitor. */
inline void lock_objects () {
    trace ("lock_objects\n");
    jvmpi->RawMonitorEnter (objects_access_lock);
}

/**  unlock the object monitor. */
inline void unlock_objects () {
    trace ("unlock_objects\n");
    jvmpi->RawMonitorExit (objects_access_lock);
}

/** Get the lock on the thread monitor. */
inline void lock_threads () {
    trace ("lock_threads\n");
    jvmpi->RawMonitorEnter (data_access_lock);    
}

/** unlock the thread monitor. */
inline void unlock_threads () {
    trace ("unlock_threads\n");
    jvmpi->RawMonitorExit (data_access_lock);
}

/** Test if jmp is currently tracing objects. */
inline int tracing_objects () {
    return object_profiling;
}

/** Test if jmp is currently tracing method calls. */
inline int tracing_methods () {
    return method_profiling;
}

/** Test if jmp is using its GUI */
inline int usingUI () {
    return doUI;
}

/** Reset the counter on all classes and method. */
void reset_counters () {
    lock_objects();
    lock_threads ();
    hash_for_each ((hash_iter_f)cls_reset_count, classes);
    hash_for_each ((hash_iter_f)method_reset_count, methods);
    unlock_threads ();
    unlock_objects();
}

/** Restore the counter to theire full values. */
void restore_counters () {
    lock_objects();
    lock_threads ();
    hash_for_each ((hash_iter_f)cls_restore_count, classes);
    hash_for_each ((hash_iter_f)method_restore_count, methods);
    unlock_threads ();
    unlock_objects();
}

/** If we are in the GTK+ thread we dont want to update status fields 
 *  since that will deadlock (trying to wait for a lock this thread is holding).
 */
static int no_gc_status = 0;

/** Run the garbage collector. */
void run_GC () {
    no_gc_status = 1;
    jvmpi->RunGC ();    
    no_gc_status = 0;
}

/** Perform a heao dump with the specified level.
 */
static void run_heap_dump_arg (int level) {
    JVMPI_HeapDumpArg heap_arg;
    heap_arg.heap_dump_level = level;
    lock_objects ();
    lock_threads ();
    jvmpi->RequestEvent (JVMPI_EVENT_HEAP_DUMP, &heap_arg);
    unlock_threads ();
    unlock_objects ();
}

void run_heap_dump () {
    run_heap_dump_arg (JVMPI_DUMP_LEVEL_2);
}

/** Ask the jvm about a specific object allocation.
 */
void get_object_alloc (jobjectID obj_id) {
    jint res = jvmpi->RequestEvent (JVMPI_EVENT_OBJECT_ALLOC, obj_id);
    if (res != JVMPI_SUCCESS)
	fprintf (stderr, "failed to get object (%p), error: %d\n", 
		 obj_id, res);
}

/** Ask the jvm about a specified class load.
 */
void get_class_load (jobjectID class_id) {
    jint res = jvmpi->RequestEvent (JVMPI_EVENT_CLASS_LOAD, class_id);
    if (res != JVMPI_SUCCESS) 
	fprintf (stderr, "failed to get class (%p), error: %d\n", 
		 class_id, res);
}

/** Load a primitive array class. */
static cls* load_array_class (char* cls_name) {
    cls* c;
    c = cls_new (cls_name, cls_name, (jobjectID)cls_name, 0, 0, NULL, 0, NULL);
    if (classes)
	hash_insert (c, classes);
    return c;
}

/** Load all of the primitive array classes. */
static void load_array_classes () {
    LP = load_array_class (L);
    ZP = load_array_class (Z);
    BP = load_array_class (B);
    CP = load_array_class (C);
    SP = load_array_class (S);
    IP = load_array_class (I);
    JP = load_array_class (J);
    FP = load_array_class (F);
    DP = load_array_class (D);
}

/** Do initial setup. */
static int setup () {
    trace ("JMP setup ()\n");
    arenas = hash_new (11, arena_hash_func, arena_cmp_func);
    classes = hash_new (2000, cls_hash_func, cls_cmp_func);
    methods = hash_new (10000, method_hash_func, method_cmp_func);
    threads = hash_new (50, jthread_hash_func, jthread_cmp_func);
    objects = hash_new (100000, obj_hash_func, obj_cmp_func);
    osp = objectstore_new ();
    load_array_classes ();    
    return 0;
}

/** Tear down all allocations we have made. 
 */
static int teardown () {
    trace ("JMP teardown ()\n");
    fprintf (stdout, _("teardown called, freeing jmp-data..\n"));
    if (objects) {
	/*fprintf (stdout, _("freeing objects: %d..\n"), hash_cardinal (objects));*/
	/* freeing the objectstore will free all objects in it.. */
	/*hash_for_each ((hash_iter_f)teardown_obj, objects); */
	objectstore_free (osp);
	hash_free (objects);
	objects = NULL;
    }
    if (threads) {
	/*fprintf (stdout, _("freeing threads: %d..\n"), hash_cardinal (threads));*/
	hash_for_each ((hash_iter_f)jthread_free, threads);
	hash_free (threads);
	threads = NULL;	
    }
    if (methods) {
	/*fprintf (stdout, _("freeing methods: %d..\n"), hash_cardinal (methods));*/
	hash_for_each ((hash_iter_f)method_free, methods);
	hash_free (methods);
	methods = NULL;
    }
    if (classes) {
	/*fprintf (stdout, _("freeing classes: %d..\n"), hash_cardinal (classes));*/
	hash_for_each ((hash_iter_f)cls_free, classes);
	hash_free (classes);
	classes = NULL;
    }
    if (arenas) {
	/*fprintf (stdout, _("freeing arenas: %d..\n"), hash_cardinal (arenas));*/
	hash_for_each ((hash_iter_f)arena_free, arenas);
	hash_free (arenas);
	arenas = NULL;
    }
    fprintf (stdout, _("done freeing jmp-data..\n"));
    
    free (dumpfile); /* NULL or malloc'ed */
    dumpfile = NULL;

    quit_ui ();
    fprintf (stdout, _("teardown complete.\n"));
    down = 1;
    return 0;
}

/** Get a class pointer from its id. */
cls* get_class (const jobjectID class_id) {
    cls c;
    cls* cp = NULL;
    cls_set_class_id (&c, class_id);
    cp = hash_search (&c, classes);
    return cp;
}

/** Get the hashtable of all loaded classes. */
hashtab* get_classes () {
    return classes;
}

/** Get an object pointer from its id. */
obj* get_object (const jobjectID obj_id) {
    obj o;
    obj* op = NULL;
    obj_set_object_id (&o, obj_id);
    op = hash_search (&o, objects);
    return op;
}

/** Get a method object pointer from its id. */
method* get_method (const jmethodID method_id) {
    method m;
    method* mp = NULL;
    method_set_method_id (&m, method_id);
    mp = hash_search (&m, methods);
    return mp;
}

/** Get the hashtable of all methods in the system. */
hashtab* get_methods () {
    return methods;
}

/** Get an jthread pointer from its id. */
jthread* get_jthread (const JNIEnv* env_id) {
    jthread t;
    jthread* tp = NULL;
    jthread_set_env_id (&t, (JNIEnv*)env_id);
    tp = hash_search (&t, threads);
    return tp;
}

/** get the timerstack for the given thread. */
timerstack* get_timerstack (JNIEnv* env_id) {
    return (timerstack*)jvmpi->GetThreadLocalStorage (env_id);    
}

hashtab* get_threads () {
    return threads;
}

/** Perform additional setup that cant be done in setup () 
 * 
 */
static void jvm_init_done (JNIEnv* env) {
    jint tstatus;
    trace ("jvm_init_done.\nContinuing with JMP setup...\n");
    
    if (doUI || (get_dump_timer () > 0)) {
	tstatus = jvmpi->CreateSystemThread("jmp-gtk", 
					    JVMPI_NORMAL_PRIORITY, 
					    gtkthread);
	if (tstatus == JNI_ERR)
	    fprintf (stdout, 
		     "GTK thread create status: %d (ok: %d, bad: %d)\n", 
		     tstatus, 
		     JNI_OK, 
		     JNI_ERR);
    }
    
    trace ("JMP setup complete...\n");    
    
}

static int run_data_dump_unlocked () {
    int ret = -1;
    if (objects && classes && methods) 
	ret = dump_data (dumpfile, classes, methods, threads);
    return ret;
}

int run_data_dump () {
    int ret = -1;
    trace ("run_data_dump ()\n");
    lock_objects ();
    ret = run_data_dump_unlocked ();
    unlock_objects ();    
    return ret;
}

void run_show_objects_alloced_by_method (method* m) {
    trace2 ("run_show_objects_alloced_by_method (%s)\n", m->jmpname);
    lock_objects (); 
    if (objects) 
	show_objects_alloced_by_method (objects, m);
    unlock_objects ();
}

void run_show_objects_alloced_by_class (cls* c) {
    trace2 ("run_show_objects_alloced_by_class (%s)\n", c->name);
    lock_objects (); 
    if (objects) 
	show_objects_alloced_by_class (objects, c);
    unlock_objects ();
}

void run_find_owners (cls* c) {
    lock_objects ();
    lock_threads ();
    show_instance_owners (objects, c);
    unlock_threads ();
    unlock_objects ();
}

cls* get_super_class (cls* c) {
    if (c->super)
	return c->super;
    jvmpi->RequestEvent (JVMPI_EVENT_OBJECT_DUMP, c->class_id);
    return c->super;
}

void check_objects (size_t cnt) {
    int diff = hash_cardinal (objects) - cnt;
    fprintf (stderr, "objects has: %u, cnt is %u (%u)\n", hash_cardinal (objects), cnt, diff);
}

static void data_reset () {
    trace ("data_reset ()\n");
}

static void disable_events () {
    /* Disable standard events */
    jvmpi->DisableEvent (JVMPI_EVENT_JVM_INIT_DONE, NULL);
    jvmpi->DisableEvent (JVMPI_EVENT_JVM_SHUT_DOWN, NULL);
    
    jvmpi->DisableEvent (JVMPI_EVENT_GC_START, NULL);
    jvmpi->DisableEvent (JVMPI_EVENT_GC_FINISH, NULL);

    jvmpi->DisableEvent (JVMPI_EVENT_ARENA_NEW, NULL);
    jvmpi->DisableEvent (JVMPI_EVENT_ARENA_DELETE, NULL);

    jvmpi->DisableEvent (JVMPI_EVENT_CLASS_LOAD, NULL);
    jvmpi->DisableEvent (JVMPI_EVENT_CLASS_UNLOAD, NULL);

    if (object_profiling) {
	/* Disable object tracing events. */
	jvmpi->DisableEvent (JVMPI_EVENT_OBJECT_ALLOC, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_OBJECT_MOVE, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_OBJECT_FREE, NULL);
    }
    
    jvmpi->DisableEvent (JVMPI_EVENT_THREAD_START, NULL);
    jvmpi->DisableEvent (JVMPI_EVENT_THREAD_END, NULL);
    
    if (method_profiling) {
	/* Disable method profiling events. */
	jvmpi->DisableEvent (JVMPI_EVENT_METHOD_ENTRY, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_METHOD_EXIT, NULL);
    }

    if (dump_enabled) {
	jvmpi->DisableEvent (JVMPI_EVENT_DATA_DUMP_REQUEST, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_DATA_RESET_REQUEST, NULL);
    }

    if (monitor_profiling) {
	jvmpi->DisableEvent (JVMPI_EVENT_MONITOR_CONTENDED_ENTER, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_MONITOR_CONTENDED_ENTERED, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_MONITOR_CONTENDED_EXIT , NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_MONITOR_WAIT, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_MONITOR_WAITED , NULL);

	/*
	jvmpi->DisableEvent (JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED, NULL);
	jvmpi->DisableEvent (JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT, NULL);
	*/
    }
}

/** The jvm is shutting down so clean up our mess.
 *  TODO dump stats to stdout/file here!
 */
static void jvm_shut_down () {
    if (down)
	return;
    disable_events ();
    lock_objects ();
    lock_threads ();
    trace ("jvm_shut_down ()\n");
    fprintf (stdout, "jvm_shut_down.\n");
    /* print statistics. */
    fprintf (stdout, "c_class_load: %ld\n", c_class_load);
    fprintf (stdout, "c_class_unload: %ld\n", c_class_unload);
    fprintf (stdout, "c_object_alloc: %ld\n", c_object_alloc);
    fprintf (stdout, "c_object_move: %ld\n", c_object_move);
    fprintf (stdout, "c_object_free: %ld\n", c_object_free);
    fprintf (stdout, "c_thread_start: %ld\n", c_thread_start);
    fprintf (stdout, "c_thread_end: %ld\n", c_thread_end);
    fprintf (stdout, "c_method_entry: %ld\n", c_method_entry);
    fprintf (stdout, "c_method_exit: %ld\n", c_method_exit);

    run_data_dump_unlocked ();
    
    teardown ();
    unlock_threads ();
    unlock_objects ();
}

/** grab the locks needed during gc.
 */
static void gc_start () {
    trace ("gc_start ()\n");
    if (!no_gc_status)
	set_jvmpi_status (_("Running garbage collection..."));
    else 
	set_status (_("Running garbage collection..."));	
    lock_objects ();
}

/** Releas the locks needed taken during gc. Show some info.
 */
static void gc_finish (jlong used_objects,
		       jlong used_object_space,
		       jlong total_object_space) {
    trace ("gc_finish ()\n");
    unlock_objects ();
    if (!no_gc_status)
	set_jvmpi_status (_("Garbage collection completed"));
    else 
	set_status (_("Garbage collection completed"));
}

/** A new arena has arrived.
 */
static void jvmpi_arena_new (jint arena_id, 
			     const char* arena_name) {
    arena* a;
    char buffer[80];
    snprintf(buffer,80,_("Arena %s created."),arena_name);
    set_jvmpi_status(buffer);
    trace3 ("arena_new: (%d, %s)\n", arena_id, arena_name);
    if (!arenas)
	return;
    lock_objects ();
    a = arena_new (arena_id, arena_name);
    if (a == NULL) {
	fprintf (stderr, "arena_new: failed to allocate arena %d, '%s'\n", arena_id, arena_name);
    } else {
	if (arenas)
	    hash_insert (a, arenas);
    }
    unlock_objects ();
}

static void object_free (jobjectID);

/** Free an object if it is in the arena that is being destroyed.. */
void free_object_if_in_arena (obj* o, jint* arena_id) {
    if (o->arena_id == *arena_id)
	object_free (o->obj_id);
}

/** An arena is destroyed, every object in arena are also deallocated so remove
 *  them (we do not get an explicit object_freed for them).
 */
static void arena_delete (jint arena_id) {
    arena *as = NULL;
    arena *ah = NULL;
    
    trace2 ("arena_delete (%d)\n", arena_id);
    as = arena_new (arena_id, "");
    ah = hash_search (as, arenas);
    arena_free (as);
    if (!ah) {
	fprintf (stderr, "arena_delete: failed to find freed arena: %d.\n", arena_id);
    } else {
	hash_remove (ah, arenas);
	arena_free (ah);
    }
    hash_for_each_with_arg ((hash_iter_fa)free_object_if_in_arena, objects, &arena_id);
}

/** A class has been loaded. 
 *  Locks will only be taken if requested = 0.
 *  Requested class_load are assumed to be locked already..
 */
static void class_load (JNIEnv* env,
			const char *class_name, 
			const char *source_name, 
			jint num_interfaces,
			jint num_methods, 
			JVMPI_Method *meths, 
			jint num_static_fields,
			JVMPI_Field *statics, 
			jint num_instance_fields, 
			JVMPI_Field *instances, 
			jobjectID class_id,
			jint requested) {

    cls* c;
    method* m;
    int i;

    trace3 ("class_load: (%s, %p ...)\n", class_name, class_id);

    if (!classes)
	return;    
    if (!requested)
	lock_objects ();
    c_class_load++;
    /* Store the class. */
    c = get_class (class_id);
    if (c == NULL) {
	c = cls_new (class_name, source_name, class_id, num_interfaces, 
		     num_static_fields, statics, 
		     num_instance_fields, instances);
	if (c == NULL) {
	    fprintf (stderr, "class_load: failed to allocate cls: %s, %s, %p.\n", 
		     class_name, source_name, class_id);
	} else {
	    hash_insert (c, classes);
	    if (!strcmp ("java/lang/Object", class_name))
		cls_print (c);			
	}

	/* and its methods */
	if (!requested)
	    lock_threads ();
	for (i = 0; i < num_methods; i++) {
	    JVMPI_Method *jmethod;
	    jmethod = meths + i;
	    m = get_method (jmethod->method_id);
	    if (m == NULL) {
		m = method_new (jmethod->method_name, jmethod->method_signature,
				jmethod->start_lineno, jmethod->end_lineno, 
				jmethod->method_id, c);
		if (m == NULL) {
		    fprintf (stderr, "class_load: failed to allocate method: %s, %p, %s, %s.\n",
			     class_name, class_id, jmethod->method_name, jmethod->method_signature);
		}
		hash_insert (m, methods);
	    }
	}
	if (!requested)
	    unlock_threads ();	    
    }

    if (!requested) {
	unlock_objects ();
    }
}

/** A class has been unloaded.
 * TODO evaluate what should happen to methods for this class, we
 * probably want to keep them. 
 */
static void class_unload (jobjectID class_id) {
    cls* cp = NULL;

    trace2 ("class_unload (%p)\n", class_id);
    c_class_unload++;
    cp = get_class (class_id);
    if (!cp) {
	fprintf (stderr, "class_unload: failed to find unloaded class: %p.\n", class_id);
    } else {
	/* dont remove it if we have objects that have not been GC:ed... */
	/*hash_remove (cp, classes);
	  cls_free (cp);*/
    }
}

static cls* get_class_pointer (jint is_array, jobjectID class_id) {
    cls* cp = NULL;
    switch (is_array) {
    case JVMPI_NORMAL_OBJECT:	
	cp = get_class (class_id);
	if (!cp) {
	    /* Ok, we dont know about this class yet, request it */
	    if (class_id != 0)
		jvmpi->RequestEvent(JVMPI_EVENT_CLASS_LOAD, class_id);
	    cp = get_class (class_id);	    
	}
	break;
    case JVMPI_CLASS:
	/* The documentation say that class_id is null always for [L... to bad.*/
	cp = LP;
	break;
    case JVMPI_BOOLEAN:
	cp = ZP;
	break;
    case JVMPI_BYTE:
	cp = BP; 
	break;
    case JVMPI_CHAR:
	cp = CP; 
	break;
    case JVMPI_SHORT:
	cp = SP; 
	break;
    case JVMPI_INT:
	cp = IP; 
	break;
    case JVMPI_LONG:
	cp = JP; 
	break;
    case JVMPI_FLOAT:
	cp = FP; 
	break;
    case JVMPI_DOUBLE:
	cp = DP; 
	break;
    } 
    return cp;
}

/** Create a timerstack for the given thread.
 */
static timerstack* setup_thread_local_storage (JNIEnv* thread_env_id) {
    timerstack* s;
    s = (timerstack*)jvmpi->GetThreadLocalStorage (thread_env_id);
    if (s)
	return s;
    s = timerstack_new (100);
    if (s == NULL) 
	fprintf (stderr, "thread_start: failed to allocate thread local stoarge.\n");
    jvmpi->SetThreadLocalStorage (thread_env_id, s);
    return s;
}

static void setup_unknown_thread (JNIEnv* env_id) { 
    /* Ok, since we have the GC disabled this will deadlock
     * if we try anything fancy like: 
     * jobjectID thread_id = jvmpi->GetThreadObject (env_id);
     * jvmpi->RequestEvent (JVMPI_EVENT_THREAD_START, thread_id);
     * 
     * We can not get stack frames for a thread that has not started.
     */

    /* we do not need to lock, we are the thread that handle this storage. */
    setup_thread_local_storage (env_id);    
}

/** A java object is beeing allocated...
 */
static void object_alloc (jint arena_id, jobjectID class_id, jint is_array,
			  jint size, jobjectID obj_id, jint requested, JNIEnv* env) {
    cls* cp = NULL;
    obj* op = NULL;
    method* mp = NULL;	    

    if (!objects)
	return;
    
    trace3 ("object_alloc (%p, %p)\n", class_id, obj_id);
    lock_objects ();
    c_object_alloc++;
    cp = get_class_pointer (is_array, class_id);
    if (cp != NULL) {
	if (method_profiling) {
	    /** We dont need to lock the thread/method since the thread that is allocing is the 
	     *  one we are in.
	     */
	    methodtime* mt;
	    timerstack* t = get_timerstack (env);
	    if (t == NULL) {
		setup_unknown_thread (env);
		t = get_timerstack (env);
	    }
	    if (t) {
		/* SUN's jvm seems evil. It creates a lot of objects before sending thread start.
		 * Calling getCallTrace on these will give 0 frames back (to bad).
		 * which means that we wont get any method here...
		 */
		if (t->top > 0) {
		    mt =  t->times + (t->top - 1);
		    mp = get_method (mt->method_id);
		} 
	    }
	}
	op = objectstore_obj_new (osp, arena_id, cp, is_array, size, obj_id, mp);
	hash_insert (op, objects);    
	cls_object_alloc (cp, size);
    } else {
	fprintf (stderr, "failed to find class(%p) for object_alloc (%p)"
		 "(probably java.lang.Class?)\n", class_id, obj_id);
    }
    
    unlock_objects ();
}

static void object_move (jint arena_id, jobjectID obj_id, 
			 jint new_arena_id, jobjectID new_obj_id) {    
    obj* op = NULL;
    obj* hop = NULL;
    cls* cp = NULL;

    trace3 ("object_move (%p, %p)\n", obj_id, new_obj_id);
    c_object_move++;
    if (!objects)
	return;
    op = get_object (obj_id);
    if (op != NULL) {
	hop = hash_remove (op, objects);
	obj_set_arena_id (op, new_arena_id);
	obj_set_object_id (op, new_obj_id);
	hash_insert (op, objects);
    }

    /* try to get class */
    cp = get_class (obj_id);   
    if (cp) {
	/* have to have a good hash table since new 
	 * classes will reuse the old object  id 
	 */
	hash_remove (cp, classes);
	cls_set_class_id (cp, new_obj_id);
	hash_insert (cp, classes);
    } 
}

static void object_free (jobjectID obj_id) {
    obj* op = NULL;
    obj* hop = NULL;
    cls* cp = NULL;
    jint os;
    
    trace2 ("object_free (%p)\n", obj_id);
    c_object_free++;
    if (!objects)
	return;
    op = get_object (obj_id);
    if (op != NULL) {  /* we ignore objects that we know nothing about. */
	hop = hash_remove (op, objects);
	cp = get_class (obj_get_class_id (hop));
	if (cp != NULL) {
	    os = obj_get_size (hop);
	    cls_object_free (cp, os);
	}
	objectstore_obj_free (osp, hop);
    } 
}

static void thread_start_nolock (char* thread_name, 
				 char* group_name,
				 char* parent_name,
				 jobjectID thread_id,
				 JNIEnv* thread_env_id, 
				 jint requested) {
    jthread* t;
    c_thread_start++;
    if (threads) {
	t = get_jthread (thread_env_id);
	if (t == NULL) {
	    t = jthread_new (thread_name, group_name, parent_name, thread_id, thread_env_id);
	    if (t == NULL) {
		fprintf (stderr, "thread_start: failed to allocate jthread: %s, %s, %s, %p, %p\n",
			 thread_name, group_name, parent_name, thread_id, thread_env_id);
	    }
	    hash_insert (t, threads);
	    setup_thread_local_storage (thread_env_id);
	}
    }
}

static void thread_start (char* thread_name, 
			  char* group_name,
			  char* parent_name,
			  jobjectID thread_id,
			  JNIEnv* thread_env_id, 
			  jint requested) {
    char buffer[80];
    snprintf (buffer,80,_("Thread %s started"),thread_name);
    set_jvmpi_status (buffer);
    trace4 ("thread_start (%s,%s,%s)\n", thread_name, group_name, parent_name);
    lock_threads ();
    thread_start_nolock (thread_name, group_name, parent_name, thread_id, thread_env_id, requested);
    unlock_threads ();
}

static void thread_end (JNIEnv* env) {
    jthread* t;
    timerstack* s;

    trace ("thread_end ()\n");
    if (!threads)
	return;
    lock_threads ();
    c_thread_end++;
    t = get_jthread (env);
    if (!t) {
	fprintf (stderr, "failed to find thread that ended: %p\n", env);
    } else {
	hash_remove (t, threads);
	jthread_free (t);
    }
    
    s = (timerstack*)jvmpi->GetThreadLocalStorage (env);
    timerstack_free (s);
    unlock_threads ();
}

/** A method is beeing called. */
static void method_entry (jmethodID method_id, JNIEnv *env) {
    timerstack* s;
    struct timeval tv;

    trace ("method_entry (...)\n");
    c_method_entry++;
    s = (timerstack*)jvmpi->GetThreadLocalStorage (env);
    if (s == NULL) 
	s = setup_thread_local_storage (env);
    gettimeofday (&tv, NULL);
    jthread_method_entry (s, method_id, &tv);
    trace ("method_entry (...) done\n");
}

/** A method has exited. 
 */
static void method_exit (jmethodID method_id, JNIEnv *env) {
    method* m;
    timerstack* s;
    struct timeval tv;
    
    trace ("method_exit (...)\n");
    c_method_exit++;
    if (!methods)
	return;
    /* try to get the method without locking. Normally it is there. */
    m = get_method (method_id);
    if (m == NULL) {
	/* ok the method was not there, retry with the lock held, it should work. */
	lock_threads ();
	m = get_method (method_id);
	/* ok, try really hard to get it... */
	if (m == NULL) {
	    jobjectID class_id;
	    cls* c;
	    jvmpi->DisableGC ();
	    class_id = jvmpi->GetMethodClass (method_id);
	    get_class_load (class_id);
	    c = get_class (class_id);
	    m = get_method (method_id);
	    jvmpi->EnableGC ();
	}
	unlock_threads (); 
    }
    if (m != NULL) {	
	gettimeofday (&tv, NULL);
	s = (timerstack*)jvmpi->GetThreadLocalStorage (env);
	if (s == NULL) 
	    s = setup_thread_local_storage (env);
    	jthread_method_exit (s, m, &tv);
    } else {
	fprintf (stdout, "method_exit: method not found (%p)\n", method_id);
    }
    trace ("method_exit (...) done\n");
}


static void heap_dump (int dump_level,
		       char *begin,
		       char *end,
		       jint num_traces,
		       JVMPI_CallTrace *traces) {
    trace4 ("heap dump dump_level: %d, begin: %p, end: %p, ", dump_level, begin, end);
    trace3 ("num_traces: %d, traces: %p\n", num_traces, traces);    
    switch (dump_level) {
    case JVMPI_DUMP_LEVEL_0:
	heap_dump_0 (dump_level, begin, end, num_traces, traces);
	break;
    case JVMPI_DUMP_LEVEL_1:
	heap_dump_1 (dump_level, begin, end, num_traces, traces);
	break;
    case JVMPI_DUMP_LEVEL_2:
	heap_dump_2 (dump_level, begin, end, num_traces, traces);
	break;
    }
}


typedef void (*contend_func)(timerstack*, obj*, struct timeval*, jthread*);

static void monitor_contend_handler (JNIEnv* env_id, jobjectID object, char* trace, contend_func func) {
    struct timeval tv;
    jthread* jt;
    obj* o;
    timerstack* s;
    /* ok, now we try real hard to get what we need. */
    jt = get_jthread (env_id);
    if (!jt) {
	jobjectID class_id;
	jvmpi->DisableGC ();
	class_id = jvmpi->GetThreadObject (env_id);
	jvmpi->RequestEvent(JVMPI_EVENT_THREAD_START, class_id);
	jvmpi->EnableGC ();
	jt = get_jthread (env_id);
    }
    o = get_object (object);
    if (!o) {
	jvmpi->RequestEvent(JVMPI_EVENT_OBJECT_ALLOC, object);
	o = get_object (object);
    }
    trace4 ("%s %s, %s\n", trace, jthread_get_thread_name (jt), cls_get_name (obj_get_class (o)));
    s = (timerstack*)jvmpi->GetThreadLocalStorage (env_id);
    if (s == NULL) 
	s = setup_thread_local_storage (env_id);
    gettimeofday (&tv, NULL);
    func (s, o, &tv, jt);
}


static void monitor_contended_enter (JNIEnv* env_id, jobjectID object) {
    monitor_contend_handler (env_id, object, "monitor contended enter", 
			     (contend_func)jthread_contenation_enter);
}

static void monitor_contended_entered (JNIEnv* env_id, jobjectID object) {
    monitor_contend_handler (env_id, object, "monitor contended entered", 
			     (contend_func)jthread_contenation_entered);
}

static inline void monitor_contended_exit (JNIEnv* env_id, jobjectID object) {
    trace ("monitor_contended_exit\n");
    /*
    jthread* jt = get_jthread (env_id);
    obj* o = get_object (object);
    fprintf (stderr, "JVMPI_EVENT_MONITOR_CONTENDED_EXIT   : env: %p => %s, obj: %p => %s\n", 
	     env_id,jthread_get_thread_name (jt),  object, cls_get_name (obj_get_class (o)));
    */
}

static void monitor_wait (JNIEnv* env_id, jobjectID object, jlong timeout) {
    trace ("monitor_wait\n");
    /*
    jthread* jt = get_jthread (env_id);
    obj* o = get_object (object);
    fprintf (stderr, "JVMPI_EVENT_MONITOR_WAIT  : env: %p => %s, obj: %p => %s, %lld\n", 
	     env_id, jthread_get_thread_name (jt), object, cls_get_name (obj_get_class (o)), timeout);
    */
}

static void monitor_waited (JNIEnv* env_id, jobjectID object, jlong timeout) {
    trace ("monitor_waited\n");
    /*
    jthread* jt = get_jthread (env_id);
    obj* o = get_object (object);
    fprintf (stderr, "JVMPI_EVENT_MONITOR_WAITED: env: %p => %s, obj: %p => %s, %lld\n", 
	     env_id, jthread_get_thread_name (jt), object, cls_get_name (obj_get_class (o)), timeout);
    */
}

/** Switch on the event_type and call the right method. 
 */
void notifyEvent(JVMPI_Event *e) {
    trace3 ("e->event_type: %d, e->env: %p: ", e->event_type, e->env_id);
    switch(e->event_type) {
    case JVMPI_EVENT_JVM_INIT_DONE:
	jvm_init_done (e->env_id);
	break;
    case JVMPI_EVENT_JVM_SHUT_DOWN:
	jvm_shut_down ();
	break;
    case JVMPI_EVENT_GC_START:
	gc_start ();
	break;
    case JVMPI_EVENT_GC_FINISH:    
	gc_finish (e->u.gc_info.used_objects, 
		   e->u.gc_info.used_object_space,
		   e->u.gc_info.total_object_space);		
	break;
    case JVMPI_EVENT_ARENA_NEW: 
	jvmpi_arena_new (e->u.new_arena.arena_id, 
			 e->u.new_arena.arena_name);
	break;
    case JVMPI_EVENT_ARENA_DELETE:
	arena_delete (e->u.delete_arena.arena_id);
	break;
    case JVMPI_EVENT_CLASS_LOAD:
    case JVMPI_EVENT_CLASS_LOAD | JVMPI_REQUESTED_EVENT:
	class_load (e->env_id,
		    e->u.class_load.class_name, e->u.class_load.source_name,
		    e->u.class_load.num_interfaces, e->u.class_load.num_methods,
		    e->u.class_load.methods, e->u.class_load.num_static_fields,
		    e->u.class_load.statics, e->u.class_load.num_instance_fields, 
		    e->u.class_load.instances, e->u.class_load.class_id, 
		    e->event_type & JVMPI_REQUESTED_EVENT);
	break;
    case JVMPI_EVENT_CLASS_UNLOAD:
	class_unload (e->u.class_unload.class_id);
	break;	
    case JVMPI_EVENT_OBJECT_ALLOC:
    case JVMPI_EVENT_OBJECT_ALLOC | JVMPI_REQUESTED_EVENT:
	object_alloc (e->u.obj_alloc.arena_id, e->u.obj_alloc.class_id, e->u.obj_alloc.is_array,
		      e->u.obj_alloc.size, e->u.obj_alloc.obj_id,
		      e->event_type & JVMPI_REQUESTED_EVENT, e->env_id);
	break;
    case JVMPI_EVENT_OBJECT_MOVE:
	object_move (e->u.obj_move.arena_id, e->u.obj_move.obj_id, 
		     e->u.obj_move.new_arena_id, e->u.obj_move.new_obj_id);
	break;
    case JVMPI_EVENT_OBJECT_FREE:
	object_free (e->u.obj_free.obj_id);	
	break;
    case JVMPI_EVENT_THREAD_START:
    case JVMPI_EVENT_THREAD_START | JVMPI_REQUESTED_EVENT:
	thread_start (e->u.thread_start.thread_name, 
		      e->u.thread_start.group_name, 
		      e->u.thread_start.parent_name,
		      e->u.thread_start.thread_id, 
		      e->u.thread_start.thread_env_id,
		      e->event_type & JVMPI_REQUESTED_EVENT);
	break;
    case JVMPI_EVENT_THREAD_END:
	thread_end (e->env_id);
	break;
    case JVMPI_EVENT_METHOD_ENTRY:
	method_entry (e->u.method.method_id, e->env_id);
	break;
    case JVMPI_EVENT_METHOD_EXIT:
	method_exit (e->u.method.method_id, e->env_id);
	break;
    case JVMPI_EVENT_DATA_DUMP_REQUEST: 
	run_data_dump ();
	break;
    case JVMPI_EVENT_DATA_RESET_REQUEST:
	data_reset ();
	break;	
    case JVMPI_EVENT_HEAP_DUMP:
    case JVMPI_EVENT_HEAP_DUMP | JVMPI_REQUESTED_EVENT:
	heap_dump (e->u.heap_dump.dump_level,
		   e->u.heap_dump.begin,
		   e->u.heap_dump.end,
		   e->u.heap_dump.num_traces,
		   e->u.heap_dump.traces);
	break;
    case JVMPI_EVENT_OBJECT_DUMP:
    case JVMPI_EVENT_OBJECT_DUMP | JVMPI_REQUESTED_EVENT:
	object_dump (e->u.object_dump.data_len,
		     e->u.object_dump.data);
	break;
    case JVMPI_EVENT_MONITOR_CONTENDED_ENTER:
	monitor_contended_enter (e->env_id, e->u.monitor.object);
	break;
    case JVMPI_EVENT_MONITOR_CONTENDED_ENTERED:
	monitor_contended_entered (e->env_id, e->u.monitor.object);
	break;
    case JVMPI_EVENT_MONITOR_CONTENDED_EXIT :
	monitor_contended_exit (e->env_id, e->u.monitor.object);
	break;
    case JVMPI_EVENT_MONITOR_WAIT:
	monitor_wait (e->env_id, e->u.monitor_wait.object, e->u.monitor_wait.timeout);
	break;
    case JVMPI_EVENT_MONITOR_WAITED :
	monitor_waited (e->env_id, e->u.monitor_wait.object, e->u.monitor_wait.timeout);
	break;

    case JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER:
	fprintf (stderr, "raw_monitor_contended_enter  :%p, %p => %s\n", 
		 e->env_id, e->u.raw_monitor.id, e->u.raw_monitor.name);
	break;
    case JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED:
	fprintf (stderr, "raw_monitor_contended_entere :d%p, %p => %s\n", 
		 e->env_id, e->u.raw_monitor.id, e->u.raw_monitor.name);
	break;
    case JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT:
	fprintf (stderr, "raw_monitor_contended_exit   :%p, %p => %s\n", 
		 e->env_id, e->u.raw_monitor.id, e->u.raw_monitor.name);
	break;
    default:
	trace ("unhandled.\n");
    }
}

static void display_help () {
    fprintf (stdout, "java -Xrunjmp[:[options]] package.Class\n");
    fprintf (stdout, "options is a comma separated list and may include:\n");
    fprintf (stdout, "help      - to show this text.\n");
    fprintf (stdout, "nomethods - to disable method profiling.\n");
    fprintf (stdout, "noobjects - to disable object profiling.\n");
    fprintf (stdout, "nomonitors - to disable monitor profiling.\n");
    fprintf (stdout, "nogui     - to run jmp without the user interface.\n");
    fprintf (stdout, "dodump    - to allow to be called with signals.\n");
    fprintf (stdout, "dumpfile=<filename> - to specify the dumpfiles.\n");
    fprintf (stdout, "dumptimer=<n> - to specify automatic dump every n:th second.\n");
    fprintf (stdout, "\nAn example may look like this:\n"); 
    fprintf (stdout, "java -Xrunjmp:nomethods,dumpfile=/tmp/jmpdump rabbit.proxy.Proxy\n"); 
}

static size_t get_size_of_option (char* option) {
    size_t t;
    char* n;
    n = strstr (option, ",");
    if (n != NULL) 
	t = n - option;
    else 
	t = strlen (option);
    return t;
}

static void parse_options (char* options) {
    char* m = NULL;

    /** Parse the options we got. */
    if (options != NULL) {
	m = strstr (options, "help");
	if (m != NULL) {
	    fprintf (stderr, "help wanted..\n");
	    display_help ();
	    exit (0);
	}

	m = strstr (options, "nomethods");
	method_profiling = (m == NULL);
	
	m = strstr (options, "noobjects");
	object_profiling = (m == NULL);
	
	m = strstr (options, "nomonitors");
	monitor_profiling = (m == NULL);

	m = strstr (options, "nogui");
	doUI = (m == NULL);
	
	m = strstr (options, "dodump");
	dump_enabled = (m != NULL);
	
	m = strstr (options, "dumpfile=");
	if (m != NULL) {
	    size_t t = get_size_of_option (m + 9);	    
	    dumpfile = malloc (t + 1);
	    strncpy (dumpfile, m + 9, t);
	    dumpfile[t] = '\0';
	}
	
	m = strstr (options, "dumptimer=");
	if (m != NULL) {
	    char *e;
	    size_t t = get_size_of_option (m + 10); 
	    long l = strtol (m + 10, &e, 0);
	    if (e != m + 10 + t) {
		fprintf (stdout, "strange dumptimer (%ld) value specified, ignored: %p, %p...\n", 
			 l, (m + 10), (m + 10 + t));
	    } else {
		 set_dump_timer (l);
	    }		
	}
    }
}

static void enable_events () {
    /* Enable standard events */
    jvmpi->EnableEvent (JVMPI_EVENT_JVM_INIT_DONE, NULL);
    jvmpi->EnableEvent (JVMPI_EVENT_JVM_SHUT_DOWN, NULL);
    
    jvmpi->EnableEvent (JVMPI_EVENT_GC_START, NULL);
    jvmpi->EnableEvent (JVMPI_EVENT_GC_FINISH, NULL);

    jvmpi->EnableEvent (JVMPI_EVENT_ARENA_NEW, NULL);
    jvmpi->EnableEvent (JVMPI_EVENT_ARENA_DELETE, NULL);

    jvmpi->EnableEvent (JVMPI_EVENT_CLASS_LOAD, NULL);
    jvmpi->EnableEvent (JVMPI_EVENT_CLASS_UNLOAD, NULL);

    if (object_profiling) {
	/* Enable object tracing events. */
	jvmpi->EnableEvent (JVMPI_EVENT_OBJECT_ALLOC, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_OBJECT_MOVE, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_OBJECT_FREE, NULL);
    }
    
    jvmpi->EnableEvent (JVMPI_EVENT_THREAD_START, NULL);
    jvmpi->EnableEvent (JVMPI_EVENT_THREAD_END, NULL);
    
    if (method_profiling) {
	/* Enable method profiling events. */
	jvmpi->EnableEvent (JVMPI_EVENT_METHOD_ENTRY, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_METHOD_EXIT, NULL);
    }

    if (dump_enabled) {
	jvmpi->EnableEvent (JVMPI_EVENT_DATA_DUMP_REQUEST, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_DATA_RESET_REQUEST, NULL);
    }

    if (monitor_profiling) {
	jvmpi->EnableEvent (JVMPI_EVENT_MONITOR_CONTENDED_ENTER, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_MONITOR_CONTENDED_ENTERED, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_MONITOR_CONTENDED_EXIT , NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_MONITOR_WAIT, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_MONITOR_WAITED , NULL);

	/*
	jvmpi->EnableEvent (JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED, NULL);
	jvmpi->EnableEvent (JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT, NULL);
	*/
    }
}

/** This method is the one that is called by the jvm on startup. 
 */
JNIEXPORT jint JNICALL JVM_OnLoad(JavaVM *jvm, char *options, void *reserved) {
    fprintf (stdout, "jmp initializing: (%s):...\n", (options ? options : ""));
    parse_options (options);
    fprintf (stdout, "    tracing objects: %s\n", (object_profiling ? "true" : "false"));
    fprintf (stdout, "    tracing methods: %s\n", (method_profiling ? "true" : "false"));
    fprintf (stdout, "    tracing monitors: %s\n", (monitor_profiling ? "true" : "false"));
    fprintf (stdout, "    showing gui: %s\n", (doUI ? "true" : "false"));
    fprintf (stdout, "    dump/reset by signal allowed: %s\n", (dump_enabled ? "true" : "false"));

    /* get jvmpi interface pointer */
    if (((*jvm)->GetEnv(jvm, (void **)&jvmpi, JVMPI_VERSION_1)) < 0) {
	fprintf (stderr, "jmp: error in obtaining jvmpi interface pointer\n");
	return JNI_ERR;
    }
    jvmpi->NotifyEvent = notifyEvent;   
    if (setup ())
	return JNI_ERR;
    
    data_access_lock = jvmpi->RawMonitorCreate("_jmp_data_access_lock");
    objects_access_lock = jvmpi->RawMonitorCreate("_jmp_objects_access_lock");

    fprintf (stderr, "jmp: Enabling localization.\n");
    /* Enable internationalization */
    setlocale (LC_ALL, "");
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
    fprintf (stderr, _("jmp: Loaded and registered correctly.\n"));

    /* initialize jvmpi interface */
    enable_events ();

    return JNI_OK;    
}
